from contextlib import AbstractAsyncContextManager
from importlib.machinery import ModuleSpec
from typing import Any, Callable, Generator, Sequence
from starlette.middleware import Middleware
from starlette.routing import BaseRoute, Router, Route, WebSocketRoute
from starlette.endpoints import HTTPEndpoint, WebSocketEndpoint
from starlette.types import ASGIApp
from typing import MutableMapping, Awaitable, Mapping
from types import ModuleType
import importlib.util
import os


class BaseRouter(Router):
    """
    A base router class for handling HTTP and WebSocket routes.

    Args:
        Router (type): The base router class.

    Attributes:
        directory (str): The directory path where the routes are located.
        entry_point (str): The entry point for the routes.
        file (dict[str, str]): A dictionary mapping file names to their types.
        class_callable (str): The name of the class callable for router or websocket routes.
        _type (str): The type of the router (http or websocket).
        redirect_slashes (bool): Whether to redirect trailing slashes in the routes.
        default (Callable[[MutableMapping[str, Any], Callable[[], Awaitable[MutableMapping[str, Any]]], Callable[[MutableMapping[str, Any]], Awaitable[None]]], Awaitable[None]] | None): The default handler for routes.
        on_startup (Sequence[Callable[[], Any]] | None): A sequence of functions to be called on startup.
        on_shutdown (Sequence[Callable[[], Any]] | None): A sequence of functions to be called on shutdown.
        lifespan (Callable[[Any], AbstractAsyncContextManager[None]] | Callable[[Any], AbstractAsyncContextManager[Mapping[str, Any]]] | None): A callable that manages the lifespan of the router.
        middleware (Sequence[Middleware] | None): A sequence of middleware functions to be applied to the routes.
    """

    def __init__(
        self,
        directory: str,
        entry_point: str,
        file: dict[str, str],
        class_callable: str = "endpoints",
        _type: str = "http",
        redirect_slashes: bool = True,
        default: (
            Callable[
                [
                    MutableMapping[str, Any],
                    Callable[[], Awaitable[MutableMapping[str, Any]]],
                    Callable[[MutableMapping[str, Any]], Awaitable[None]],
                ],
                Awaitable[None],
            ]
            | None
        ) = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: (
            Callable[[Any], AbstractAsyncContextManager[None]]
            | Callable[[Any], AbstractAsyncContextManager[Mapping[str, Any]]]
            | None
        ) = None,
        *,
        middleware: Sequence[Middleware] | None = None,
    ) -> None:

        self._router_type = _type
        self._callable = entry_point
        self._class_callable = class_callable
        self._directory = directory
        self._file = file
        super().__init__(
            self.generate_routes(),
            redirect_slashes,
            default,
            on_startup,
            on_shutdown,
            lifespan,
            middleware=middleware,
        )

    @property
    def directory(self) -> str:
        """Returns the directory associated with the router.

        Returns:
            str: The directory path.
        """
        return self._directory

    @property
    def callable(self) -> str:
        return self._callable

    @property
    def router_type(self) -> str:
        """Returns the type of the router.

        Returns:
            str: The type of the router.
        """
        return self._router_type

    def generate_routes(self) -> list[Route | WebSocketRoute]:
        """Generates routes based on the files in the directory.

        Returns:
            list[Route | WebSocketRoute]: A list of routes generated from the files.
        """
        router: list[Route | WebSocketRoute] = []

        for root, file in self.walk():
            module_name: str = file.split(".")[0]
            module_path: str = os.path.join(root, file)

            module: ModuleType = self.load_module(module_name, module_path)

            route_path: str = self.create_route_path(self.directory, root)

            if self._file[file] in ["router", "websocket"]:
                router.append(
                    self.create_route(
                        route_path,
                        getattr(module, self._class_callable),
                    )
                )
            else:
                router.append(
                    self.create_route(
                        route_path,
                        getattr(module, self.callable),
                        methods=[module_name],
                    )
                )

        return router

    def walk(self) -> Generator[tuple[str, str], Any, None]:
        """Recursively walks through the directory and yields tuples of root and file names.

        Yields:
            Generator[tuple[str, str], Any, None]: A generator that yields tuples of root and file names.
        """
        for root, _, files in os.walk(self.directory):
            for file in files:
                if file in self._file:
                    yield root, file

    def load_module(self, module_name: str, module_path: str) -> ModuleType:
        """
        Load a module from a file location.

        Args:
            module_name (str): The name of the module.
            module_path (str): The path to the module file.

        Raises:
            ModuleNotFoundError: If the module is not found.
            AttributeError: If the callable is not found in the module.

        Returns:
            ModuleType: The loaded module.
        """
        spec: ModuleSpec | None = importlib.util.spec_from_file_location(
            module_name, module_path
        )
        if not spec:
            raise ModuleNotFoundError("Module not found")
        module: ModuleType = importlib.util.module_from_spec(spec)
        if not spec.loader:
            raise ModuleNotFoundError("Module not found")
        spec.loader.exec_module(module)
        if not hasattr(module, self.callable) and not hasattr(
            module, self._class_callable
        ):
            raise AttributeError(f"Callable {module_name} not found in module")
        return module

    def create_route_path(self, directory: str, root: str) -> str:
        """
        Create a route path by replacing the directory in the root path and performing some string replacements.

        Args:
            directory (str): The directory to be replaced in the root path.
            root (str): The root path.

        Returns:
            str: The modified route path.
        """
        return (
            root.replace(directory, "")
            .replace("\\", "/")
            .replace("[", "{")
            .replace("]", "}")
            + "/"
        )

    def create_route(
        self,
        route_path: str,
        obj: Callable | type[HTTPEndpoint] | type[WebSocketEndpoint],
        methods: list[str] = ["GET"],
    ) -> Route | WebSocketRoute:
        """
        Create a route with the given path and endpoint.

        Args:
            route_path (str): The path of the route.
            endpoint (Callable): The endpoint of the route.

        Returns:
            BaseRoute: The created route.
        """
        if issubclass(type(obj), WebSocketEndpoint):
            return WebSocketRoute(route_path, obj, name=obj.__name__)

        if issubclass(type(obj), HTTPEndpoint):
            return Route(route_path, obj, methods=methods)

        if self.router_type == "http":
            return Route(route_path, obj, methods=methods)

        if self.router_type == "websocket":
            return WebSocketRoute(route_path, obj, name=obj.__name__)

        raise ValueError("Invalid router type")


class HTTPRouter(BaseRouter):
    """A router class for handling HTTP requests.

    Args:
        BaseRouter (type): The base router class.

    Attributes:
        directory (str): The directory where the router files are located.
        callable (str): The name of the callable function in the router files.
        class_callable (str): The name of the class containing the router files.
        file (dict[str, str]): A dictionary mapping file names to HTTP methods.
        _type (str): The type of router (in this case, "http").
        redirect_slashes (bool): Whether to redirect trailing slashes in URLs.
        default (Callable): The default handler for requests that don't match any route.
        on_startup (Sequence[Callable]): A sequence of functions to run on startup.
        on_shutdown (Sequence[Callable]): A sequence of functions to run on shutdown.
        lifespan (Callable): A function that returns an async context manager for the router's lifespan.
        middleware (Sequence[Middleware]): A sequence of middleware functions to apply to requests.

    """

    def __init__(
        self,
        directory: str,
        callable: str = "endpoint",
        class_callable: str = "Endpoints",
        file: dict[str, str] = {},
        _type: str = "http",
        redirect_slashes: bool = True,
        default: (
            Callable[
                [
                    MutableMapping[str, Any],
                    Callable[[], Awaitable[MutableMapping[str, Any]]],
                    Callable[[MutableMapping[str, Any]], Awaitable[None]],
                ],
                Awaitable[None],
            ]
            | None
        ) = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: (
            Callable[[Any], AbstractAsyncContextManager[None]]
            | Callable[[Any], AbstractAsyncContextManager[Mapping[str, Any]]]
            | None
        ) = None,
        *,
        middleware: Sequence[Middleware] | None = None,
    ) -> None:

        if not file:
            file = {
                "get.py": "GET",
                "post.py": "POST",
                "put.py": "PUT",
                "delete.py": "DELETE",
                "patch.py": "PATCH",
                "head.py": "HEAD",
                "options.py": "OPTIONS",
                "router.py": "router",
            }

        super().__init__(
            directory,
            callable,
            file,
            class_callable,
            _type,
            redirect_slashes,
            default,
            on_startup,
            on_shutdown,
            lifespan,
            middleware=middleware,
        )


class WebsocketRouter(BaseRouter):
    """
    A router for handling WebSocket connections.

    This router is responsible for handling WebSocket connections in a web application.

    Args:
        BaseRouter (type): The base router class to inherit from.

    Attributes:
        directory (str): The directory where the router is located.
        entry_point (str): The entry point for the router.
        class_callable (str): The name of the class that contains the endpoints.
        file (dict[str, str]): A dictionary mapping file names to file types.
        _type (str): The type of router.
        redirect_slashes (bool): Whether to redirect trailing slashes in URLs.
        default (Callable[[MutableMapping[str, Any], Callable[[], Awaitable[MutableMapping[str, Any]]], Callable[[MutableMapping[str, Any]], Awaitable[None]]], Awaitable[None]] | None): The default handler for requests.
        on_startup (Sequence[Callable[[], Any]] | None): A sequence of functions to run on application startup.
        on_shutdown (Sequence[Callable[[], Any]] | None): A sequence of functions to run on application shutdown.
        lifespan (Callable[[Any], AbstractAsyncContextManager[None]] | Callable[[Any], AbstractAsyncContextManager[Mapping[str, Any]]] | None): A function that returns an async context manager for managing the lifespan of the application.
        middleware (Sequence[Middleware] | None): A sequence of middleware functions to apply to requests.

    """

    def __init__(
        self,
        directory: str,
        entry_point: str = "endpoint",
        class_callable: str = "Endpoints",
        file: dict[str, str] = {},
        _type: str = "websocket",
        redirect_slashes: bool = True,
        default: (
            Callable[
                [
                    MutableMapping[str, Any],
                    Callable[[], Awaitable[MutableMapping[str, Any]]],
                    Callable[[MutableMapping[str, Any]], Awaitable[None]],
                ],
                Awaitable[None],
            ]
            | None
        ) = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: (
            Callable[[Any], AbstractAsyncContextManager[None]]
            | Callable[[Any], AbstractAsyncContextManager[Mapping[str, Any]]]
            | None
        ) = None,
        *,
        middleware: Sequence[Middleware] | None = None,
    ) -> None:

        if not file:
            file = {"websocket.py": "websocket"}
        super().__init__(
            directory,
            entry_point,
            file,
            class_callable,
            _type,
            redirect_slashes,
            default,
            on_startup,
            on_shutdown,
            lifespan,
            middleware=middleware,
        )



from starlette.middleware.base import BaseHTTPMiddleware
from jose import jwt
from jose.exceptions import ExpiredSignatureError, JWTError
from starlette.responses import JSONResponse
import base64


class BearerTokenMiddleware(BaseHTTPMiddleware):
    """
    Middleware for handling bearer token authentication.

    Args:
        app (ASGIApp): The ASGI application to wrap with the middleware.
        secret_key (str): The secret key used for token verification.
        algorithm (str): The algorithm used for token verification.
        expires_in (int, optional): The expiration time for tokens in seconds. Defaults to 3600.
    """

    def __init__(self, app, secret_key: str, algorithm: str, expires_in: int = 3600):
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.expires_in = expires_in
        super().__init__(app)

    async def dispatch(self, request, call_next):
        """Dispatches the request and performs authentication.

        Args:
            request (Request): The incoming request object.
            call_next (Callable): The next middleware or route handler.

        Returns:
            Response: The response returned by the next middleware or route handler.
        """
        bearer, token = request.headers.get("Authorization", "").split(" ")
        if bearer.lower() != "bearer":
            return JSONResponse({"detail": "Invalid token type"}, status_code=401)
        try:
            payload = jwt.decode(token, self.secret_key, algorithms=[self.algorithm])
        except ExpiredSignatureError:
            return JSONResponse({"detail": "Token has expired"}, status_code=401)
        except JWTError:
            return JSONResponse({"detail": "Invalid token"}, status_code=401)
        request.state.user = payload
        return await call_next(request)


class BasicTokenMiddleware(BaseHTTPMiddleware):
    """Middleware for handling basic token authentication.

    This middleware extracts the username and password from the Authorization header,
    validates the token type, and sets the user information in the request state.

    Args:
        app (ASGIApp): The ASGI application to wrap with this middleware.
    """

    def __init__(self, app):
        super().__init__(app)

    async def dispatch(self, request, call_next):
        """Dispatch method that handles the request.

        Args:
            request (Request): The incoming request.
            call_next (Callable): The next middleware or application to call.

        Returns:
            Response: The response returned by the next middleware or application.
        """
        basic, token = request.headers.get("Authorization", "").split(" ")
        if basic.lower() != "basic":
            return JSONResponse({"detail": "Invalid token type"}, status_code=401)
        username, _, password = base64.b64decode(token).decode().partition(":")
        request.state.user = {"username": username, "password": password}
        return await call_next(request)



from starlette.requests import Request
from typing import Any
from pypox._types import BodyDict
from pypox.processing.validators.base import Validator


class JSONValidator(Validator):

    async def validate(self, _type: type, request: Request) -> None:
        if _type not in [BodyDict]:
            return None
        return _type.__supertype__(await request.json())



from starlette.requests import Request
from typing import Any
from pypox._types import BodyForm
from pypox.processing.validators.base import Validator


class FormValidator(Validator):

    async def validate(self, _type: type, request: Request) -> None:
        if not _type in [BodyForm]:
            return None
        return _type.__supertype__(await request.form())



from starlette.requests import Request
from typing import Any
from pypox._types import PathStr, PathInt, PathFloat, PathBool
from pypox.processing.validators.base import Validator


class PathValidator(Validator):

    def __init__(self, name: str, _type: type) -> None:
        super().__init__(name, _type)

    async def validate(self, _type: type, request: Request) -> None:
        if _type not in [PathStr, PathInt, PathFloat, PathBool]:
            return None
        if self._name.replace("_", "-") in request.path_params:
            value = request.path_params.get(self._name.replace("_", "-"))
        else:
            value = request.path_params.get(self._name)
        if not value:
            return None
        return _type.__supertype__(value)



from typing import Any, Callable, Mapping, Sequence, Type
from jinja2.defaults import (
    BLOCK_END_STRING,
    BLOCK_START_STRING,
    COMMENT_END_STRING,
    COMMENT_START_STRING,
    KEEP_TRAILING_NEWLINE,
    LSTRIP_BLOCKS,
    TRIM_BLOCKS,
    VARIABLE_END_STRING,
    VARIABLE_START_STRING,
)
from jinja2.ext import Extension
from starlette.background import BackgroundTask
from starlette.requests import Request
from pydantic import BaseModel, field_validator, Field
from pypox.processing.base import HeaderValidator
from jinja2 import BaseLoader, BytecodeCache, Environment, FileSystemLoader, Undefined
from starlette.responses import HTMLResponse


class HTMXValidator(HeaderValidator):

    def __init__(self, name: str, _type: Type) -> None:
        super().__init__(name, _type)

    async def validate(self, _type: Type, request: Request) -> None:
        if _type not in [HTMXHeaders, HTMXResponseHeaders]:
            return None
        return _type(**request.headers)


class HTMXHeaders(BaseModel):

    boosted: str = Field(default="false", alias="hx-boosted")
    current_url: str = Field(default="", alias="hx-current-url")
    history_restored: str = Field(default="false", alias="hx-history-restored")
    prompt: str = Field(default="false", alias="hx-prompt")
    request: str = Field(default="false", alias="hx-request")
    target: str = Field(default="", alias="hx-target")
    trigger_name: str = Field(default="", alias="hx-trigger-name")
    trigger: str = Field(default="", alias="hx-trigger")

    @field_validator("request")
    def validate_request(cls, value):
        if value:
            return value
        else:
            raise ValueError("Request must be True")


class HTMXResponseHeaders(BaseModel):
    location: str = Field(
        default="",
        alias="hx-location",
    )
    push_url: str = Field(
        default="",
        alias="hx-push-url",
    )
    redirect: str = Field(
        default="",
        alias="hx-redirect",
    )
    refresh: str = Field(
        default="",
        alias="hx-refresh",
    )
    replace_url: str = Field(
        default="",
        alias="hx-replace-url",
    )
    reswap: str = Field(
        default="",
        alias="hx-reswap",
    )
    retarget: str = Field(
        default="",
        alias="hx-retarget",
    )
    reselect: str = Field(
        default="",
        alias="hx-reselect",
    )
    trigger: str = Field(
        default="",
        alias="hx-trigger",
    )
    trigger_after_settle: str = Field(
        default="",
        alias="hx-trigger-after-settle",
    )
    trigger_after_swap: str = Field(
        default="",
        alias="hx-trigger-after-swap",
    )



from starlette.requests import Request
from typing import Any
from pypox._types import CookieStr, CookieInt, CookieFloat, CookieBool
from pypox.processing.validators.base import Validator


class CookieValidator(Validator):

    async def validate(self, _type: type, request: Request) -> None:
        if _type not in [CookieStr, CookieInt, CookieFloat, CookieBool]:
            return None
        if self._name.replace("_", "-") in request.cookies:
            value = request.cookies.get(self._name.replace("_", "-"))
        else:
            value = request.cookies.get(self._name)
        if not value:
            return None
        return _type.__supertype__(value)



from abc import ABC
from starlette.requests import Request


class Validator(ABC):

    def __init__(self, name: str, _type: type) -> None:
        self._name = name
        self._type = _type

    @property
    def name(self) -> str:
        return self._name

    async def validate(self, _type: type, request: Request) -> None:
        raise NotImplementedError("validate method must be implemented")

    async def __call__(self, request: Request) -> dict:
        value = await self.validate(self._type, request)
        if not value:
            return {}
        else:
            return {self._name: value}



from typing import NewType


QueryStr = NewType("QueryStr", str)
QueryInt = NewType("QueryInt", int)
QueryFloat = NewType("QueryFloat", float)
QueryBool = NewType("QueryBool", bool)
PathStr = NewType("PathStr", str)
PathInt = NewType("PathInt", int)
PathFloat = NewType("PathFloat", float)
PathBool = NewType("PathBool", bool)
HeaderStr = NewType("HeaderStr", str)
HeaderInt = NewType("HeaderInt", int)
HeaderFloat = NewType("HeaderFloat", float)
HeaderBool = NewType("HeaderBool", bool)
CookieStr = NewType("CookieStr", str)
CookieInt = NewType("CookieInt", int)
CookieFloat = NewType("CookieFloat", float)
CookieBool = NewType("CookieBool", bool)
BodyDict = NewType("BodyDict", dict)
BodyForm = NewType("BodyForm", dict)



from functools import wraps
import inspect
from typing import Any, Callable, Awaitable
from starlette.requests import Request
from starlette.responses import Response
import asyncio
from pypox.processing.validators.base import Validator
from pypox.processing.validators.form import FormValidator
from pypox.processing.validators.json import JSONValidator
from pypox.processing.validators.query import QueryValidator
from pypox.processing.validators.path import PathValidator
from pypox.processing.validators.header import HeaderValidator
from pypox.processing.validators.cookies import CookieValidator
from pypox.processing.validators.htmx import HTMXValidator


def processor(
    validators: list = [],
) -> Callable:
    """Decorator function that adds validation to a request handler function.

    Args:
        validators (list, optional): A list of additional validators to apply. Defaults to [].

    Returns:
        Callable: A decorated request handler function.
    """

    def decorator(
        func: Callable,
    ) -> Callable[[Request], Awaitable[Response | Response]]:

        @wraps(func)
        async def wrapper(request: Request) -> Response:
            params = await PypoxProcessor(
                func,
                [
                    QueryValidator,
                    PathValidator,
                    HeaderValidator,
                    CookieValidator,
                    JSONValidator,
                    FormValidator,
                ]
                + validators,
            ).validate(
                request,
            )
            if inspect.iscoroutinefunction(func):
                return await func(**params)
            else:
                return func(**params)

        return wrapper

    return decorator


class PypoxProcessor:
    """A class representing a Pypox processor.

    This class is responsible for processing requests and validating parameters.

    Attributes:
        _validators (list): A list of validators to be applied to the parameters.
        _func (Callable): The function to be executed for processing the request.

    """

    def __init__(
        self,
        func: Callable[[Any], Awaitable[Response] | Response],
        validators: list = [],
    ) -> None:
        self._validators = validators
        self._func = func

    async def validate(self, request: Request) -> dict:
        """Validates the request and returns the parameters.

        This method validates the given request by iterating through the parameters
        of the associated function and applying the registered validators. It returns
        a dictionary of parameters that have been validated.

        Args:
            request (Request): The request object to be validated.

        Returns:
            Any: A dictionary containing the validated parameters.

        """
        params = {}
        async for name, parameter in self.iterate_params(self._func):
            if parameter.annotation == Request:
                params[name] = request
                break
            for validator in self._validators:
                validator_obj: Validator = validator(name, parameter.annotation)
                data = await validator_obj(request)
                if data:
                    params.update(data)
                    break
        return params

    async def iterate_params(
        self, func: Callable[[Any], Awaitable[Response] | Response]
    ):
        for name, parameter in inspect.signature(func).parameters.items():
            yield name, parameter
            await asyncio.sleep(0)



import os
from typing import (
    Any,
    Awaitable,
    Callable,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    TypeVar,
    ParamSpec,
)
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route, Router, Mount, BaseRoute
from starlette.types import ExceptionHandler, Lifespan, Receive, Scope, Send
from pypox.processing.base import PypoxProcessor, processor
from pypox.router import BaseRouter
from pypox.openapi.main import OpenAPI, Info, License


class Pypox(Starlette):
    """A web application framework based on Starlette.

    Args:
        Starlette (type): The base class for the Pypox application.

    Attributes:
        processor_func (list[BaseProcessor] | None): The list of processor functions.
        conventions (list[BaseConvention]): The list of conventions.
        debug (bool): Flag indicating whether debug mode is enabled.
        middleware (Sequence[Middleware] | None): The middleware stack.
        exception_handlers (Mapping[Any, ExceptionHandler] | None): The exception handlers.
        on_startup (Sequence[Callable[[], Any]] | None): The startup functions.
        on_shutdown (Sequence[Callable[[], Any]] | None): The shutdown functions.
        lifespan (Lifespan | None): The lifespan of the application.
    """

    def __init__(
        self,
        conventions: list[BaseRouter] | None = None,
        debug: bool = False,
        middleware: Sequence[Middleware] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: Lifespan | None = None,
        open_api_version: str = "3.0.3",
        info: Info = Info(title="Pypox", version="0.0.1", license=License(name="MIT")),
        license: License = License(name="MIT"),
        validators: list[Any] = [],
    ) -> None:
        self._openapi_version = open_api_version
        self._info = info
        self._license = license
        self._validators = validators
        routes: list[BaseRoute] = []
        if conventions:
            for convention in conventions:
                endpoints = convention.generate_routes()
                for endpoint in endpoints:
                    endpoint.endpoint = PypoxProcessor(
                        endpoint.endpoint, validators
                    ).validate
                routes.extend(endpoints)
        super().__init__(
            debug,
            routes,
            middleware,
            exception_handlers,
            on_startup,
            on_shutdown,
            lifespan,
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["openapi"] = {
            "openapi": self._openapi_version,
            "info": self._info,
            "license": self._license,
        }
        return await super().__call__(scope, receive, send)

    def add_route(
        self,
        path: str,
        route: Callable,
        methods: List[str] | None = None,
        name: str | None = None,
        include_in_schema: bool = True,
        validators: (
            List[Callable[[Request], Awaitable[Response] | Response]] | None
        ) = None,
    ) -> None:
        """Add a new route to the application.

        Args:
            path (str): The URL path for the route.
            route (Callable): The function or method to be executed when the route is accessed.
            methods (List[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.
            validators (List[Callable[[Request], Awaitable[Response]  |  Response]]  |  None, optional):
                A list of validators to be applied to the request before executing the route. Defaults to None.
        """

        self.routes.append(
            Route(
                path,
                processor(validators or [])(route),
                methods=methods,
                name=name,
                include_in_schema=include_in_schema,
            )
        )


class PypoxHTMX(BaseRouter):
    """A class representing the PypoxHTMX application.

    Args:
        BaseRouter (type): The base router class.

    Attributes:
        _openapi_version (str): The OpenAPI version.
        _info (Info): The information about the application.
        _license (License): The license information.

    """

    def __init__(
        self,
        directory: str | None = None,
        middleware: Sequence[Middleware] | None = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: Lifespan | None = None,
        open_api_version: str = "3.0.3",
        info: Info = Info(title="Pypox", version="0.0.1", license=License(name="MIT")),
        license: License = License(name="MIT"),
    ) -> None:
        self._openapi_version = open_api_version
        self._info = info
        self._license = license
        super().__init__(
            directory=directory or "", entry_point="page", file={"page.py": "GET"}
        )
        if self._directory:
            self._router = Starlette(
                routes=self.generate_routes(),
                middleware=middleware,
                on_startup=on_startup,
                on_shutdown=on_shutdown,
                lifespan=lifespan,
            )

    async def __call__(
        self,
        scope: MutableMapping[str, Any],
        receive: Callable[[], Awaitable[MutableMapping[str, Any]]],
        send: Callable[[MutableMapping[str, Any]], Awaitable[None]],
    ) -> None:
        """Handle incoming requests.

        Args:
            scope (MutableMapping[str, Any]): The request scope.
            receive (Callable[[], Awaitable[MutableMapping[str, Any]]]): The receive function.
            send (Callable[[MutableMapping[str, Any]], Awaitable[None]]): The send function.

        Returns:
            None: No return value.

        """
        if hasattr(self, "_router"):
            return await self._router(scope, receive, send)
        else:
            return await super().__call__(scope, receive, send)
